import asyncio
import calendar
import time

INITIAL_DELAY_SECONDS = 3
BACK_OFF_MULTIPLIER = 2
MAX_DELAY_SECONDS = 5 * 60
RESET_DELAY_AFTER_SECONDS = 10 * 60


def retry(logger=None):
    def wrapper(fp):
        async def retry_loop(*args, **kwargs):
            last_failed_time = -1
            while True:
                try:
                    await fp(*args, **kwargs)
                    logger.debug("@retry decorated function %s exited, ending retry loop", fp.__name__)
                    break
                except Exception as ex:
                    now = calendar.timegm(time.gmtime())
                    if last_failed_time == -1 or (now - last_failed_time) > RESET_DELAY_AFTER_SECONDS:
                        delay_seconds = INITIAL_DELAY_SECONDS
                    else:
                        delay_seconds = delay_seconds * BACK_OFF_MULTIPLIER
                    if delay_seconds > MAX_DELAY_SECONDS:
                        delay_seconds = MAX_DELAY_SECONDS
                    last_failed_time = now
                    logger.error("%s failed: %s, retrying in %d seconds", fp.__name__, ex,
                                 delay_seconds, exc_info=True)
                    await asyncio.sleep(delay_seconds)

        return retry_loop

    return wrapper


if __name__ == "__main__":
    import logging
    logging.basicConfig(level=logging.DEBUG)
    logger = logging.getLogger("retry-test")

    @retry(logger=logger)
    async def main():
        raise Exception("blah")

    asyncio.run(main())



from __future__ import annotations
import datetime
import inspect
import json
import random
import uuid
from dataclasses import dataclass, field
from datetime import timedelta
from typing import Callable, List, Type, Dict, Tuple, Any
from uuid import uuid4

from .activity_method import RetryParameters, ActivityOptions
from .api.common.v1 import WorkflowType, WorkflowExecution, SearchAttributes, Memo
from .api.enums.v1 import WorkflowIdReusePolicy, HistoryEventFilterType, EventType
from .api.history.v1 import WorkflowExecutionTerminatedEventAttributes
from .api.query.v1 import WorkflowQuery
from .api.taskqueue.v1 import TaskQueue
from .api.workflowservice.v1 import StartWorkflowExecutionRequest, GetWorkflowExecutionHistoryRequest, \
    QueryWorkflowRequest, QueryWorkflowResponse, SignalWorkflowExecutionRequest, WorkflowServiceStub

from .constants import DEFAULT_SOCKET_TIMEOUT_SECONDS
from .converter import DataConverter, DEFAULT_DATA_CONVERTER_INSTANCE, get_fn_ret_type_hints, get_fn_args_type_hints
from .errors import QueryFailedError
from .exception_handling import deserialize_exception
from .exceptions import WorkflowFailureException, ActivityFailureException, QueryRejectedException, \
    QueryFailureException, WorkflowOperationException
from .service_helpers import create_workflow_service, get_identity


class Workflow:

    @staticmethod
    def new_activity_stub(activities_cls, retry_parameters: RetryParameters = None, activity_options: ActivityOptions = None):
        from .decision_loop import ITask
        task: ITask = ITask.current()
        assert task
        cls = activities_cls()
        cls._decision_context = task.decider.decision_context
        cls._retry_parameters = retry_parameters  # type: ignore
        cls._activity_options = activity_options
        return cls

    @staticmethod
    def new_untyped_activity_stub(retry_parameters: RetryParameters = None,
                                  activity_options: ActivityOptions = None):
        from .decision_loop import ITask
        from .activity_method import UntypedActivityStub
        task: ITask = ITask.current()
        assert task
        cls = UntypedActivityStub()
        cls._decision_context = task.decider.decision_context
        cls._retry_parameters = retry_parameters  # type: ignore
        cls._activity_options = activity_options
        return cls

    @staticmethod
    async def await_till(c: Callable, timeout_seconds: int = 0) -> bool:
        from .decision_loop import ITask
        task: ITask = ITask.current()
        assert task
        return await task.await_till(c, timeout_seconds)

    @staticmethod
    async def sleep(seconds: int):
        from .decision_loop import ITask
        task: ITask = ITask.current()
        await task.decider.decision_context.schedule_timer(seconds)

    @staticmethod
    def current_time_millis() -> int:
        from .decision_loop import ITask
        task: ITask = ITask.current()
        return int(task.decider.decision_context.current_time_millis().timestamp() * 1000)

    @staticmethod
    def now() -> datetime.datetime:
        from .decision_loop import ITask
        task: ITask = ITask.current()
        return task.decider.decision_context.current_time_millis()

    @staticmethod
    def random_uuid() -> uuid.UUID:
        from .decision_loop import ITask
        task: ITask = ITask.current()
        return task.decider.decision_context.random_uuid()

    @staticmethod
    def new_random() -> random.Random:
        from .decision_loop import ITask
        task: ITask = ITask.current()
        return task.decider.decision_context.new_random()

    @staticmethod
    def get_version(change_id: str, min_supported: int, max_supported: int):
        from .decision_loop import ITask
        from .decision_loop import DecisionContext
        task: ITask = ITask.current()
        decision_context: DecisionContext = task.decider.decision_context
        return decision_context.get_version(change_id, min_supported, max_supported)

    @staticmethod
    def get_logger(name):
        from .decision_loop import ITask
        task: ITask = ITask.current()
        return task.decider.decision_context.get_logger(name)

    @staticmethod
    def get_workflow_id():
        from .decision_loop import ITask
        task: ITask = ITask.current()
        return task.decider.workflow_execution.workflow_id

    @staticmethod
    def get_run_id():
        from .decision_loop import ITask
        task: ITask = ITask.current()
        return task.decider.workflow_execution.run_id


class WorkflowStub:
    pass


@dataclass
class WorkflowExecutionContext:
    workflow_type: str
    workflow_execution: WorkflowExecution
    workflow_method_instance: WorkflowMethod


@dataclass
class WorkflowClient:
    service: WorkflowServiceStub
    namespace: str
    options: WorkflowClientOptions
    data_converter: DataConverter

    @classmethod
    def new_client(cls, host: str = "localhost", port: int = 7233, namespace: str = "",
                   options: WorkflowClientOptions = None, timeout: int = DEFAULT_SOCKET_TIMEOUT_SECONDS,
                   data_converter: DataConverter = DEFAULT_DATA_CONVERTER_INSTANCE) -> WorkflowClient:
        service = create_workflow_service(host, port, timeout=timeout)
        return cls(service=service, namespace=namespace, options=options, data_converter=data_converter)

    @classmethod
    async def start(cls, stub_fn: Callable, *args) -> WorkflowExecutionContext:
        stub = stub_fn.__self__  # type: ignore
        client = stub._workflow_client  # type: ignore
        assert client is not None
        method = stub_fn._workflow_method  # type: ignore
        assert method is not None
        options = stub._workflow_options  # type: ignore
        return await exec_workflow(client, method, args,
                             workflow_options=options, stub_instance=stub)

    def new_workflow_stub(self, cls: Type, workflow_options: WorkflowOptions = None):
        if not workflow_options:
            workflow_options = WorkflowOptions()
        attrs: Dict[str, Any] = {}
        attrs["_workflow_client"] = self
        attrs["_workflow_options"] = workflow_options
        for name, fn in inspect.getmembers(cls, inspect.isfunction):
            if hasattr(fn, "_workflow_method"):
                attrs[name] = get_workflow_stub_fn(fn._workflow_method)
            elif hasattr(fn, "_signal_method"):
                attrs[name] = get_signal_stub_fn(fn._signal_method)
            elif hasattr(fn, "_query_method"):
                attrs[name] = get_query_stub_fn(fn._query_method)
        stub_cls = type(cls.__name__, (WorkflowStub,), attrs)
        return stub_cls()

    def new_workflow_stub_from_workflow_id(self, cls: Type, workflow_id: str):
        """
        Use it to send signals or queries to a running workflow.
        Do not call workflow methods on it
        """
        stub_instance = self.new_workflow_stub(cls)
        execution = WorkflowExecution(workflow_id=workflow_id, run_id=None)
        stub_instance._execution = execution
        return stub_instance

    async def wait_for_close(self, context: WorkflowExecutionContext) -> object:
        return await self.wait_for_close_with_workflow_id(workflow_id=context.workflow_execution.workflow_id,
                                                    run_id=context.workflow_execution.run_id,
                                                    workflow_method_instance=context.workflow_method_instance)

    async def wait_for_close_with_workflow_id(self, workflow_id: str, run_id: str = None,
                                              workflow_method_instance: WorkflowMethod = None,
                                              workflow_fn: Callable = None):
        if not workflow_method_instance and workflow_fn:
            if not hasattr(workflow_fn, "_workflow_method"):
                raise Exception("workflow_fn is not a valid workflow stub function")
            workflow_method_instance = workflow_fn._workflow_method
        while True:
            history_request = create_close_history_event_request(self, workflow_id, run_id)
            history_response = await self.service.get_workflow_execution_history(request=history_request)
            if not history_response.history.events:
                continue
            history_event = history_response.history.events[0]
            if history_event.event_type == EventType.EVENT_TYPE_WORKFLOW_EXECUTION_COMPLETED:
                completed_attributes = history_event.workflow_execution_completed_event_attributes
                type_hints = []
                if workflow_method_instance:
                    type_hints = workflow_method_instance._ret_type_hints
                payloads: List[object] = self.data_converter.from_payloads(completed_attributes.result,
                                                                           type_hints=type_hints)
                return payloads[0]
            elif history_event.event_type == EventType.EVENT_TYPE_WORKFLOW_EXECUTION_FAILED:
                failed_attributes = history_event.workflow_execution_failed_event_attributes
                exception = deserialize_exception(failed_attributes.failure)
                raise exception
                """
                TODO: Do we need any of the logic here?
                if failed_attributes.reason == "WorkflowFailureException":
                    exception = deserialize_exception(failed_attributes.details)
                    if isinstance(exception, ActivityFailureException):
                        exception.set_cause()
                    workflow_execution = WorkflowExecution(workflow_id=workflow_id, run_id=run_id)
                    raise WorkflowFailureException(workflow_type=workflow_type,
                                                   execution=workflow_execution) from exception
                else:
                    details: Dict = json.loads(failed_attributes.details)
                    detail_message = details.get("detailMessage", "")
                    raise WorkflowExecutionFailedException(failed_attributes.reason, details=details,
                                                           detail_message=detail_message)
                """
            elif history_event.event_type == EventType.EVENT_TYPE_WORKFLOW_EXECUTION_TIMED_OUT:
                raise WorkflowExecutionTimedOutException()
            elif history_event.event_type == EventType.EVENT_TYPE_WORKFLOW_EXECUTION_TERMINATED:
                terminated_attributes  = history_event.workflow_execution_terminated_event_attributes
                raise WorkflowExecutionTerminatedException(reason=terminated_attributes.reason,
                                                           details=terminated_attributes.details,
                                                           identity=terminated_attributes.identity)
            elif history_event.event_type == EventType.EVENT_TYPE_WORKFLOW_EXECUTION_CANCELED:
                raise WorkflowExecutionCanceledException()
            else:
                raise Exception("Unexpected history close event: " + str(history_event))

    def new_activity_completion_client(self):
        return ActivityCompletionClient(self)

    def close(self):
        self.service.channel.close()


async def exec_workflow(workflow_client: WorkflowClient, wm: WorkflowMethod, args, workflow_options: WorkflowOptions = None,
                  stub_instance: object = None) -> WorkflowExecutionContext:
    start_request = create_start_workflow_request(workflow_client, wm, args, workflow_options)
    start_response = await workflow_client.service.start_workflow_execution(request=start_request)
    execution = WorkflowExecution(workflow_id=start_request.workflow_id, run_id=start_response.run_id)
    stub_instance._execution = execution  # type: ignore
    return WorkflowExecutionContext(workflow_type=wm._name, workflow_execution=execution, workflow_method_instance=wm)


async def exec_workflow_sync(workflow_client: WorkflowClient, wm: WorkflowMethod, args: List,
                       workflow_options: WorkflowOptions = None, stub_instance: object = None):
    execution_context: WorkflowExecutionContext = await exec_workflow(workflow_client, wm, args,
                                                                workflow_options=workflow_options,
                                                                stub_instance=stub_instance)
    return await workflow_client.wait_for_close(execution_context)


async def exec_signal(workflow_client: WorkflowClient, sm: SignalMethod, args, stub_instance: object = None):
    assert stub_instance._execution  # type: ignore
    request = SignalWorkflowExecutionRequest()
    request.workflow_execution = stub_instance._execution  # type: ignore
    request.signal_name = sm.name
    request.input = workflow_client.data_converter.to_payloads(args)
    request.namespace = workflow_client.namespace
    response = await workflow_client.service.signal_workflow_execution(request=request)


async def exec_query(workflow_client: WorkflowClient, qm: QueryMethod, args, stub_instance: object = None):
    assert stub_instance._execution  # type: ignore
    request = QueryWorkflowRequest()
    request.execution = stub_instance._execution  # type: ignore
    request.query = WorkflowQuery()
    request.query.query_type = qm.name
    request.query.query_args = workflow_client.data_converter.to_payloads(args)
    request.namespace = workflow_client.namespace
    response: QueryWorkflowResponse = await workflow_client.service.query_workflow(request=request)
    """
    TODO: Do we need to bring back any of this error handling
    if err:
        if isinstance(err, QueryFailedError):
            cause = deserialize_exception(err.message)
            raise QueryFailureException(query_type=qm.name, execution=stub_instance._execution) from cause
        elif isinstance(err, Exception):
            raise err
        else:
            raise Exception(err)
    """
    if response.query_rejected:
        raise QueryRejectedException(response.query_rejected.status)
    return workflow_client.data_converter.from_payloads(response.query_result, qm.ret_type_hints)[0]


def create_memo(m: Dict[str, object], data_converter: DataConverter) -> Memo:
    memo = Memo();
    for k, v in m.items():
        memo.fields[k] = data_converter.to_payload(v)
    return memo


def create_search_attributes(s: Dict[str, object], data_converter: DataConverter) -> SearchAttributes:
    search_attributes = SearchAttributes()
    for k, v in s.items():
        search_attributes.indexed_fields[k] = data_converter.to_payload(v)
    return search_attributes


def create_start_workflow_request(workflow_client: WorkflowClient, wm: WorkflowMethod,
                                  args: List, workflow_options: WorkflowOptions = None) -> StartWorkflowExecutionRequest:
    start_request = StartWorkflowExecutionRequest()
    start_request.namespace = workflow_client.namespace
    start_request.workflow_id = wm._workflow_id if wm._workflow_id else str(uuid4())
    start_request.workflow_type = WorkflowType()
    start_request.workflow_type.name = wm._name
    start_request.task_queue = TaskQueue()
    start_request.task_queue.name = wm._task_queue
    start_request.input = workflow_client.data_converter.to_payloads(args)

    start_request.workflow_execution_timeout = wm._workflow_execution_timeout
    start_request.workflow_run_timeout = wm._workflow_run_timeout
    start_request.workflow_task_timeout = wm._workflow_task_timeout

    start_request.identity = get_identity()
    start_request.workflow_id_reuse_policy = wm._workflow_id_reuse_policy
    start_request.request_id = str(uuid4())
    start_request.cron_schedule = wm._cron_schedule if wm._cron_schedule else None

    if wm._memo:
        start_request.memo = create_memo(wm._memo, workflow_client.data_converter)
    if wm._search_attributes:
        start_request.search_attributes = create_search_attributes(wm._search_attributes,
                                                                   workflow_client.data_converter)

    if workflow_options:
        if workflow_options.workflow_id:
            start_request.workflow_id = workflow_options.workflow_id
        if workflow_options.workflow_id_reuse_policy is not None:
            start_request.workflow_id_reuse_policy = workflow_options.workflow_id_reuse_policy
        if workflow_options.workflow_run_timeout is not None:
            start_request.workflow_run_timeout = workflow_options.workflow_run_timeout
        if workflow_options.workflow_execution_timeout is not None:
            start_request.workflow_execution_timeout = workflow_options.workflow_execution_timeout
        if workflow_options.workflow_task_timeout is not None:
            start_request.workflow_task_timeout = workflow_options.workflow_task_timeout
        if workflow_options.task_queue:
            start_request.task_queue = workflow_options.task_queue
        if workflow_options.cron_schedule:
            start_request.cron_schedule = workflow_options.cron_schedule
        if workflow_options.memo is not None:
            start_request.memo = create_memo(workflow_options.memo, workflow_client.data_converter)
        if workflow_options.search_attributes is not None:
            start_request.search_attributes = create_search_attributes(workflow_options.search_attributes,
                                                                       workflow_client.data_converter)

    return start_request


def create_close_history_event_request(workflow_client: WorkflowClient, workflow_id: str,
                                       run_id: str) -> GetWorkflowExecutionHistoryRequest:
    history_request = GetWorkflowExecutionHistoryRequest()
    history_request.namespace = workflow_client.namespace
    history_request.execution = WorkflowExecution()
    history_request.execution.workflow_id = workflow_id
    history_request.execution.run_id = run_id
    history_request.wait_for_new_event = True
    history_request.history_event_filter_type = HistoryEventFilterType.HISTORY_EVENT_FILTER_TYPE_CLOSE_EVENT
    return history_request


def get_workflow_method_name(method):
    return "::".join(method.__qualname__.split(".")[-2:])


def get_workflow_stub_fn(wm: WorkflowMethod):
    async def workflow_stub_fn(self, *args):
        assert self._workflow_client is not None
        return await exec_workflow_sync(self._workflow_client, wm, args,
                                  workflow_options=self._workflow_options, stub_instance=self)

    workflow_stub_fn._workflow_method = wm  # type: ignore
    return workflow_stub_fn


def get_signal_stub_fn(sm: SignalMethod):
    async def signal_stub_fn(self, *args):
        assert self._workflow_client is not None
        return await exec_signal(self._workflow_client, sm, args, stub_instance=self)

    signal_stub_fn._signal_method = sm  # type: ignore
    return signal_stub_fn


def get_query_stub_fn(qm: QueryMethod):
    async def query_stub_fn(self, *args):
        assert self._workflow_client is not None
        return await exec_query(self._workflow_client, qm, args, stub_instance=self)

    query_stub_fn._query_method = qm  # type: ignore
    return query_stub_fn


@dataclass
class WorkflowMethod(object):
    _name: str = None
    _workflow_id: str = None
    _workflow_id_reuse_policy: WorkflowIdReusePolicy = None
    _task_queue: str = None
    _cron_schedule: str = None
    _workflow_execution_timeout: timedelta = None
    _workflow_run_timeout: timedelta = None
    _workflow_task_timeout: timedelta = None
    _memo: Dict[str, object] = None
    _search_attributes: Dict[str, object] = None
    _ret_type_hints: List[type] = None


def workflow_method(func=None,
                    name=None,
                    workflow_id=None,
                    workflow_id_reuse_policy=WorkflowIdReusePolicy.WORKFLOW_ID_REUSE_POLICY_ALLOW_DUPLICATE_FAILED_ONLY,
                    workflow_execution_timeout=timedelta(seconds=0),
                    workflow_run_timeout=timedelta(seconds=0),
                    workflow_task_timeout=timedelta(seconds=60),  # default on the server is 10s
                    task_queue=None,
                    memo: Dict[str, object] = None,
                    search_attributes: Dict[str, object] = None):
    def wrapper(fn):
        if not hasattr(fn, "_workflow_method"):
            fn._workflow_method = WorkflowMethod()
        fn._workflow_method._name = name if name else get_workflow_method_name(fn)
        fn._workflow_method._workflow_id = workflow_id
        fn._workflow_method._workflow_id_reuse_policy = workflow_id_reuse_policy
        fn._workflow_method._task_queue = task_queue
        fn._workflow_method._workflow_execution_timeout = workflow_execution_timeout
        fn._workflow_method._workflow_run_timeout = workflow_run_timeout
        fn._workflow_method._workflow_task_timeout = workflow_task_timeout
        fn._workflow_method._memo = memo
        fn._workflow_method._search_attributes = search_attributes
        fn._workflow_method._ret_type_hints = get_fn_ret_type_hints(fn)
        return fn

    if func and inspect.isfunction(func):
        return wrapper(func)
    else:
        return wrapper


@dataclass
class QueryMethod:
    name: str = None
    ret_type_hints: List[type] = None


def query_method(func=None, name: str = None):
    def wrapper(fn):
        fn._query_method = QueryMethod()
        fn._query_method.name = name if name else get_workflow_method_name(fn)
        fn._query_method.ret_type_hints = get_fn_ret_type_hints(fn)
        return fn

    if func and inspect.isfunction(func):
        return wrapper(func)
    else:
        return wrapper


@dataclass
class SignalMethod:
    name: str = None


def signal_method(func=None, name: str = None):
    def wrapper(fn):
        fn._signal_method = SignalMethod()
        fn._signal_method.name = name if name else get_workflow_method_name(fn)
        return fn

    if func and inspect.isfunction(func):
        return wrapper(func)
    else:
        return wrapper


def cron_schedule(value):
    def wrapper(fn):
        if not hasattr(fn, "_workflow_method"):
            fn._workflow_method = WorkflowMethod()
        fn._workflow_method._cron_schedule = value
        return fn

    return wrapper


@dataclass
class WorkflowClientOptions:
    pass


@dataclass
class RetryOptions:
    # TODO: Support for retry_options
    pass


@dataclass
class ContextPropagator:
    # TODO: Support for ContextPropagator
    pass


@dataclass
class WorkflowOptions:
    workflow_id: str = None
    workflow_id_reuse_policy: WorkflowIdReusePolicy = None
    workflow_run_timeout: timedelta = None
    workflow_execution_timeout: timedelta = None
    workflow_task_timeout: timedelta = None
    task_queue: str = None
    retry_options: RetryOptions = None
    cron_schedule: str = None
    memo: Dict[str, object] = None
    search_attributes: Dict[str, object] = None
    # TODO: What is this for?
    # context_propagators = List[ContextPropagator] = None


@dataclass
class WorkflowExecutionFailedException(Exception):
    reason: str
    details: Dict[str, Any]
    detail_message: str

    def __str__(self) -> str:
        cause = self.details.get("cause")
        if isinstance(cause, dict):
            return f"{cause['class']}: {cause['detailMessage']}"
        else:
            return f"{self.reason}: {self.detail_message}"


@dataclass
class WorkflowExecutionTimedOutException(Exception):
    pass


@dataclass
class WorkflowExecutionCanceledException(Exception):
    pass


@dataclass
class WorkflowExecutionTerminatedException(Exception):
    reason: str
    details: object
    identity: str

    def __str__(self) -> str:
        return self.reason


from .activity import ActivityCompletionClient



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: temporal/api/query/v1/message.proto
# plugin: python-betterproto
from dataclasses import dataclass

import betterproto

from temporal.api.common import v1 as v1common
from temporal.api.enums import v1 as v1enums


@dataclass
class WorkflowQuery(betterproto.Message):
    query_type: str = betterproto.string_field(1)
    query_args: v1common.Payloads = betterproto.message_field(2)


@dataclass
class WorkflowQueryResult(betterproto.Message):
    result_type: v1enums.QueryResultType = betterproto.enum_field(1)
    answer: v1common.Payloads = betterproto.message_field(2)
    error_message: str = betterproto.string_field(3)


@dataclass
class QueryRejected(betterproto.Message):
    status: v1enums.WorkflowExecutionStatus = betterproto.enum_field(1)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: temporal/api/version/v1/message.proto
# plugin: python-betterproto
from dataclasses import dataclass

import betterproto


@dataclass
class SupportedSDKVersions(betterproto.Message):
    """SupportedSDKVersions contains the support versions for SDK."""

    go_sdk: str = betterproto.string_field(1)
    java_sdk: str = betterproto.string_field(2)


@dataclass
class WorkerVersionInfo(betterproto.Message):
    implementation: str = betterproto.string_field(1)
    feature_version: str = betterproto.string_field(2)



from dataclasses import dataclass, field
from typing import List

from temporal.api.workflowservice.v1 import WorkflowServiceStub
from temporal.worker import Worker, WorkerOptions
from temporal.workflow import WorkflowClient


@dataclass
class WorkerFactoryOptions:
    pass


@dataclass
class WorkerFactory:
    client: WorkflowClient
    namespace: str = None
    options: WorkerFactoryOptions = None
    workers: List[Worker] = field(default_factory=list)

    def new_worker(self, task_queue: str, worker_options: WorkerOptions = None) -> Worker:
        worker = Worker(client=self.client, namespace=self.namespace, task_queue=task_queue, options=worker_options)
        self.workers.append(worker)
        return worker

    def start(self):
        for worker in self.workers:
            worker.start()



import logging
from dataclasses import dataclass, field
from typing import Callable, Dict, Any, Union
from datetime import datetime, tzinfo, timedelta

import pytz

from .api.command.v1 import StartTimerCommandAttributes
from .api.common.v1 import Payloads, Payload
from .api.history.v1 import TimerFiredEventAttributes, HistoryEvent, TimerCanceledEventAttributes
from .decision_loop import ReplayDecider, DecisionContext
from .exceptions import CancellationException
from .marker import MarkerHandler, MarkerInterface, MarkerResult
from .util import OpenRequestInfo
from . import DEFAULT_VERSION

logger = logging.getLogger(__name__)

SIDE_EFFECT_MARKER_NAME = "SideEffect"
MUTABLE_SIDE_EFFECT_MARKER_NAME = "MutableSideEffect"
VERSION_MARKER_NAME = "Version"
LOCAL_ACTIVITY_MARKER_NAME = "LocalActivity"


@dataclass
class ClockDecisionContext:
    decider: ReplayDecider
    decision_context: DecisionContext
    scheduled_timers: Dict[int, OpenRequestInfo] = field(default_factory=dict)
    replay_current_time_milliseconds: datetime = datetime.fromtimestamp(0, tz=pytz.UTC)
    replaying: bool = True
    version_handler: MarkerHandler = None

    def __post_init__(self):
        self.version_handler = MarkerHandler(self.decision_context, VERSION_MARKER_NAME)

    def set_replay_current_time_milliseconds(self, s: datetime):
        self.replay_current_time_milliseconds = s

    def current_time_millis(self) -> datetime:
        return self.replay_current_time_milliseconds

    def create_timer(self, delay_seconds: int, callback: Callable):
        if delay_seconds < 0:
            raise Exception("Negative delay seconds: " + str(delay_seconds))
        if delay_seconds == 0:
            callback(None)
            return None
        firing_time = (self.current_time_millis().timestamp() * 1000) + delay_seconds * 1000
        context = OpenRequestInfo(user_context=firing_time)
        timer = StartTimerCommandAttributes()
        timer.start_to_fire_timeout = timedelta(seconds=delay_seconds)
        timer.timer_id = str(self.decider.get_and_increment_next_id())
        start_event_id: int = self.decider.start_timer(timer)
        context.completion_handle = lambda ctx, e: callback(e)  # type: ignore
        self.scheduled_timers[start_event_id] = context
        return TimerCancellationHandler(start_event_id=start_event_id, clock_decision_context=self)

    def is_replaying(self):
        return self.replaying

    def set_replaying(self, replaying):
        self.replaying = replaying

    def timer_cancelled(self, start_event_id: int, reason: Exception):
        scheduled: OpenRequestInfo = self.scheduled_timers.pop(start_event_id, None)
        if not scheduled:
            return
        callback = scheduled.completion_handle
        exception = CancellationException("Cancelled by request")
        exception.init_cause(reason)
        callback(None, exception)

    def handle_timer_fired(self, attributes: TimerFiredEventAttributes):
        started_event_id: int = attributes.started_event_id
        if self.decider.handle_timer_closed(attributes):
            scheduled = self.scheduled_timers.pop(started_event_id, None)
            if scheduled:
                callback = scheduled.completion_handle
                callback(None, None)

    def handle_timer_canceled(self, event: HistoryEvent):
        attributes: TimerCanceledEventAttributes = event.timer_canceled_event_attributes
        started_event_id: int = attributes.started_event_id
        if self.decider.handle_timer_canceled(event):
            self.timer_cancelled(started_event_id, None)

    def get_version(self, change_id: str, min_supported: int, max_supported: int) -> int:
        def func():
            d: Dict[str, Payloads] = {"VERSION": self.decider.worker.client.data_converter.to_payloads([max_supported])}
            return d

        result: Dict[str, Payloads] = self.version_handler.handle(change_id, func)
        if result is None:
            result = {"VERSION": self.decider.worker.client.data_converter.to_payloads([DEFAULT_VERSION])}
            self.version_handler.set_data(change_id, result)
            self.version_handler.mark_replayed(change_id)  # so that we don't ever emit a MarkerRecorded for this

        version: int = self.decider.worker.client.data_converter.from_payloads(result["VERSION"])[0]  # type: ignore
        self.validate_version(change_id, version, min_supported, max_supported)
        return version

    def validate_version(self, change_id: str, version: int, min_supported: int, max_supported: int):
        if version < min_supported or version > max_supported:
            raise Exception(f"Version {version} of changeID {change_id} is not supported. "
                            f"Supported version is between {min_supported} and {max_supported}.")

    def handle_marker_recorded(self, event: HistoryEvent):
        """
        Will be executed more than once for the same event.
        """
        attributes = event.marker_recorded_event_attributes
        name: str = attributes.marker_name
        if SIDE_EFFECT_MARKER_NAME == name:
            # TODO
            # sideEffectResults.put(event.getEventId(), attributes.getDetails());
            pass
        elif LOCAL_ACTIVITY_MARKER_NAME == name:
            # TODO
            # handleLocalActivityMarker(attributes);
            pass
        elif VERSION_MARKER_NAME == name:
            marker_data = MarkerInterface.from_event_attributes(attributes)
            change_id: str = marker_data.get_id()
            data: Dict[str, Payloads] = marker_data.get_data()
            self.version_handler.mutable_marker_results[change_id] = MarkerResult(data=data)
        elif MUTABLE_SIDE_EFFECT_MARKER_NAME != name:
            # TODO
            # if (log.isWarnEnabled()) {
            #       log.warn("Unexpected marker: " + event);
            # }
            pass


@dataclass
class TimerCancellationHandler:
    start_event_id: int
    clock_decision_context: ClockDecisionContext

    def accept(self, reason: Exception):
        self.clock_decision_context.decider.cancel_timer(self.start_event_id,
                                                         lambda: self.clock_decision_context.timer_cancelled(
                                                             self.start_event_id, reason))
